import os
import shutil


class DiskFileManager:
    """Reader and Writer for disk files.


    target
      A directory on the disk. If it doesn't exist, it will be created.

    replace
      If the ``target`` directory exists, should it be completely replaced
      or simply appended to ?
    """


    def __init__(self, target, replace=False):
        self.target = target
        if replace and os.path.exists(target):
            shutil.rmtree(target)
        if not os.path.exists(target):
            os.makedirs(target)

    @staticmethod
    def list_files(directory):
        """Return the list of all file objects in the directory."""
        path = directory._path
        return [] if not os.path.exists(path) else [
            name for name in os.listdir(path)
            if os.path.isfile(os.path.join(path, name))
        ]

    @staticmethod
    def list_dirs(directory):
        """Return the list of all directory objects in the directory."""
        path = directory._path
        return [] if not os.path.exists(path) else [
            name for name in os.listdir(path)
            if os.path.isdir(os.path.join(path, name))
        ]

    @staticmethod
    def read(fileobject, mode="r"):
        """Return the entire content of a file. The mode can be 'r' or 'rb'."""
        with open(fileobject._path, mode=mode) as f:
            result = f.read()
        return result

    @staticmethod
    def write(fileobject, content, mode="a"):
        """Write the content (str, bytes) to the given file object."""
        with open(fileobject._path, mode=mode) as f:
            f.write(content)

    @staticmethod
    def delete(target):
        """Delete the file on disk."""
        if target._is_dir:
            shutil.rmtree(target._path)
        else:
            os.remove(target._path)

    def create(self, target, replace=False):
        """Create a new, empty file or directory on disk."""
        path = target._path
        if replace and os.path.exists(path):
            self.delete(target)
        if replace or (not os.path.exists(path)):
            if target._is_dir:
                os.mkdir(path)
            else:
                with open(path, "w") as f:
                    pass

    @staticmethod
    def join_paths(*paths):
        """Join paths in a system/independent way -- actually os.path.join."""
        return os.path.join(*paths)

    @staticmethod
    def close():
        """This method does nothing for DiskFileManagers."""
        pass

    def open(self, fileobject, mode="a"):
        """Open a file on disk at the location given by the file object."""
        return open(fileobject._path, mode=mode)



import os
import sys
import zipfile
import re
from collections import defaultdict

PYTHON3 = (sys.version_info[0] == 3)

if PYTHON3:
    from io import StringIO, BytesIO
    StringBytesIO = BytesIO

else:
    from StringIO import StringIO
    BytesIO = StringIO
    StringBytesIO = StringIO

EMPTY_ZIP_BYTES = b'PK\x05\x06' + 18 * b'\x00'

class ZipFileManager:
    """Reader and Writer of Zip files.

    Parameters
    ----------

    path
      Path to a zip archive to be read or written to.

    source
      Either a string/byte representing zipped data, or a file-like object
      connected to zipped data.

    replace
      In case the provided ``path`` is pointing to an already-existing file,
      should it be erased or appended to ?
    """
    # The Zipfile manager manages at the same time files already in the zip
    # archive when it was created, and files left uncompressed in memory.
    # The uncompressed files in memory are flushed into the archive upon
    # closing of the manager, with the ``.close`` method.

    def __init__(self, path=None, source=None, replace=False):
        self.path = "." if path is None else path
        if path == "@memory": # VIRTUAL ZIP FROM SCRATCH
            self.source = StringBytesIO()
            self.writer = zipfile.ZipFile(self.source, "a",
                                          compression=zipfile.ZIP_DEFLATED)
            self.reader = zipfile.ZipFile(StringBytesIO(EMPTY_ZIP_BYTES), "r")
        elif path is not None:  # ON DISK ZIP
            self.source = path
            if replace or not os.path.exists(path):
                with open(self.source, "wb") as f:
                    f.write(EMPTY_ZIP_BYTES)
            self.writer = zipfile.ZipFile(self.source, "a",
                                          compression=zipfile.ZIP_DEFLATED)
            self.reader = zipfile.ZipFile(self.source, "r")
        else: # VIRTUAL ZIP FROM EXISTING DATA
            self.source = source
            if isinstance(self.source, (str, bytes)):
                self.source = StringBytesIO(source)
            self.writer = zipfile.ZipFile(self.source, "a",
                                          compression=zipfile.ZIP_DEFLATED)
            self.reader = zipfile.ZipFile(self.source, "r")
        self.files_data = defaultdict(lambda *a: StringBytesIO())

    def relative_path(self, target):
        path = target._path[len(self.path)+1:]
        if target._is_dir and path != "":
            path += "/"
        return path


    def list_files(self, directory):
        path = self.relative_path(directory)
        matches = [
            re.match(r"%s([^/]*)$" % path, name)
            for name in self.reader.namelist()
        ]
        files = set([
            match.groups()[0]
            for match in matches
            if match is not None
            and (match.groups()[0] != "")
        ])
        return files


    def list_dirs(self, directory):
        path = self.relative_path(directory)
        matches = [
            re.match(r"%s([^/]*)/" % path, name)
            for name in self.reader.namelist()
        ]
        dirs = set([
            match.groups()[0]
            for match in matches
            if match is not None
            and (match.groups()[0] != "")
        ])
        return dirs


    def read(self, fileobject, mode="r"):
        path = self.relative_path(fileobject).strip("/")
        if path in self.files_data:
            return self.files_data[path].getvalue()
        else:
            return self.reader.read(path)

    def write(self, fileobject, content, mode="w"):
        path = self.relative_path(fileobject)
        if self.path_exists_in_file(fileobject):
            raise NotImplementedError(
                "Rewriting a file already zipped is not currently supported. "
                "It may actually not even be possible, or in an inelegant way."
            )
        if mode in ("w", "wb"): # i.e. not append
            self.files_data.pop(path, None) # overwrite if exists!
        if not isinstance(content, bytes):
            content = content.encode("utf-8")
        self.files_data[path].write(content)

    def delete(self, directory):
        raise NotImplementedError(
            "Deleting/modifying/overwriting an already-zipped file "
            "is not currently supported. "
            "It may actually not even be possible, or in a very inelegant way."
        )

    def create(self, directory, replace=False):
        if self.path_exists_in_file(directory) and replace:
            self.delete(directory)
        # TODO: I don't know how to create an empty dir in a zip but right
        # now it doesn't really matter because the directories are created
        # the moment we create a file whose address in this directory.

    def path_exists_in_file(self, directory):
        return self.relative_path(directory) in self.reader.namelist()


    @staticmethod
    def join_paths(*paths):
        return "/".join(*paths)

    def close(self):
        for path, data in self.files_data.items():
            self.writer.writestr(path, data.getvalue())
        self.writer.close()
        if hasattr(self.source, "getvalue"):
            return self.source.getvalue()

    def open(self, fileobject, mode="a"):
        path = self.relative_path(fileobject)
        if mode in ("r", "rb"):
            if path in self.files_data:
                return StringBytesIO(self.files_data.getvalue())
            else:
                return StringBytesIO(self.read(fileobject, mode=mode))
        else:
            if mode == 'w' and path not in self.files_data:
                self.files_data[path] = StringIO()
            elif mode == 'wb' and path not in self.files_data:
                self.files_data[path] = BytesIO()
            return self.files_data[path]



import os
import sys
from flametree import (file_tree, DiskFileManager, ZipFileManager)

PYTHON3 = (sys.version_info[0] == 3)


zip_path = "test.zip"
test_dir = "testdir"
ALL_FILES = set(["bla.txt", "bli.txt", "blu.txt", "Readme.md"])

def test_directory(tmpdir):
    # CREATE AND POPULATE A DIRECTORY FROM SCRATCH
    dir_path = os.path.join(str(tmpdir), "test_dir")
    root = file_tree(dir_path)
    assert root._file_manager.__class__ == DiskFileManager
    root._file("Readme.md").write("This is a test zip")
    root._dir("texts")._dir("shorts")._file("bla.txt").write("bla bla bla")
    root.texts.shorts._file("bli.txt").write("bli bli bli")
    root.texts.shorts._file("blu.txt").write("blu blu blu")

    # TEST REPLACE BEHAVIOR
    root._dir("trash")._file("bla.txt").write("bla bla bla")
    root._dir("trash")._file("blu.txt").write("blu blu blu")
    print (root._tree_view())
    assert root.trash._filenames == ["blu.txt"]

    root.trash._delete()
    root._dir("trash")._file("bla.txt").write("bla bla bla")
    root._dir("trash", replace=False)._file("blu.txt").write("blu blu blu")
    assert set(root.trash._filenames) == set(["bla.txt", "blu.txt"])


    # READ AN EXISTING DIRECTORY
    root = file_tree(dir_path)
    assert set([f._name for f in root._all_files]) == ALL_FILES

    # APPEND TO AN EXISTING DIRECTORY
    root._dir("newdir")._file("new_file.txt").write("I am a new file")
    print("ah ah ah ah ah ah")
    root = file_tree(dir_path)
    print("ojojjoojojojojoj")
    assert set([f._name for f in root._all_files]) == \
        ALL_FILES.union(set(["new_file.txt"]))

    # TEST DELETION
    path = root.newdir.new_file_txt._path
    assert os.path.exists(path)
    root.newdir.new_file_txt.delete()
    assert not os.path.exists(path)
    assert not any([f._path == path for f in root.newdir._files])



def test_zip(tmpdir):
    # CREATE AND POPULATE A ZIP FROM SCRATCH, IN MEMORY
    root = file_tree("@memory")
    assert root._file_manager.__class__ == ZipFileManager
    root._file("Readme.md").write("This is a test zip")
    root._dir("texts")._dir("shorts")._file("bla.txt").write("bla bla bla")
    root.texts.shorts._file("bli.txt").write("bli bli bli")
    root.texts.shorts._file("blu.txt").write("blu blu blu")

    data = root._close()

    # READ A ZIP FROM DATA IN MEMORY
    root = file_tree(data)
    assert set([f._name for f in root._all_files]) == ALL_FILES

    # APPEND TO A CREATED ZIP IN MEMORY
    root._dir("newdir")._file("new_file.txt").write("I am a new file")
    new_data = root._close()

    # VERIFY THAT THE DATA RECEIVED CAN BE WRITTEN TO A VALID ZIP

    zip_path = os.path.join(str(tmpdir), "test.zip")
    with open(zip_path, ("wb" if PYTHON3 else "w")) as f:
        f.write(new_data)
    root = file_tree(zip_path)

    # READ ZIP FROM DISK, APPEND TO IT

    assert set([f._name for f in root._all_files]) == \
        ALL_FILES.union(set(["new_file.txt"]))
    root.newdir._file("new_file2.txt").write("I am another new file")
    root._close()

    # Final test
    root = file_tree(zip_path)
    assert set([f._name for f in root._all_files]) == \
        ALL_FILES.union(set(["new_file.txt", "new_file2.txt"]))

def test_file_tree(tmpdir):
    """Assert that the directory function dispatches as expected."""

    root = file_tree(os.path.join(str(tmpdir), "my_folder/"))
    assert root._file_manager.__class__ == DiskFileManager

    root = file_tree(os.path.join(str(tmpdir), "my_archive.zip"))
    assert root._file_manager.__class__ == ZipFileManager

    root = file_tree("@memory")
    assert root._file_manager.__class__ == ZipFileManager
    root._file("test.txt").write("bla bla bla bla bla bla bla bla")
    data = root._close()

    root = file_tree(data)
    assert root._file_manager.__class__ == ZipFileManager
    assert [f._name for f in root._all_files] == ["test.txt"]

def test_matplotlib_writer(tmpdir):
    zip_path = os.path.join(str(tmpdir), "archive.zip")
    folder_path = os.path.join(str(tmpdir), "test_folder")
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        print("Not running test_matplotlib_writer: Matplotlib not available.")
        return True
    fig, ax = plt.subplots(1)
    with file_tree(zip_path) as zip_root:
        fig_dir = zip_root._dir("figures", replace=False)
        fig.savefig(fig_dir._file("fig.png"), format="png")
        fig.savefig(fig_dir._file("fig.pdf").open("wb"), format="pdf")
    with file_tree(folder_path) as root:
        fig_dir = root._dir("figures", replace=False)
        fig.savefig(fig_dir._file("fig.png"), format="png")#
        fig.savefig(fig_dir._file("fig.pdf").open("wb"), format="pdf")
    assert (set([f._name for f in root._all_files]) ==
            set(["fig.png", "fig.pdf"]))
    assert os.path.exists(os.path.join(folder_path, "figures", "fig.pdf"))
